//
// General purpose CLIP display program.
// Reads all pictures as integers because that is the form in which
// CLIP stores them.
// John Robinson 1995-2005.
//
#include <iomanip>
#include <iostream>
#include "stdio.h"
#include "math.h"
#include "picture.h"
	

// MAGBLOCK is the length of the size of the pel squares in the inspector
// picture.
const int MAGBLOCK = 42;
// NUMBINS is the number of bins in the histograms displayed in the
// inspector picture
const int NUMBINS = 64;


// Make the histograms of colour components and draw them as graphs in
// the inspector picture (here called hists)
void makehists(colour_picture& in, colour_picture& hists) {
	int nr = hists.nrows();
	int nc = hists.ncols();
	int startrow = MAGBLOCK*8;	// Because this is below inspector
	int inmin = in.min();
	int inmax = in.max();
	int inrange = inmax - inmin + 1;
	int hist[4][NUMBINS];
	int comp, i, j;
	int numcomps = 4;
	// if (in.ismonochrome())
	// 	numcomps = 1;
	for (comp = 0; comp < numcomps; comp++) {
	    for (i = 0; i < NUMBINS; i++) {
		hist[comp][i] = 0;
		}
	    }
	int *pg, *pr, *pb;
	for (i = 0; i < in.nrows(); i++) {
		pg = in.green[i];
		pr = in.red[i];
		pb = in.blue[i];
		j = in.ncols();
		while(j--) {
			int g, r, b;
			++hist[1][(((g=*pg++)-inmin)*(NUMBINS-1))/inrange];
			++hist[2][(((r=*pr++)-inmin)*(NUMBINS-1))/inrange];
			++hist[3][(((b=*pb++)-inmin)*(NUMBINS-1))/inrange];
			++hist[0][(((g+r+b)/3-inmin)*(NUMBINS-1))/inrange];
			}
		}
	// Find global maximum value
	int histmax = 0;
	for (comp = 0; comp < numcomps; comp++) {
	    for (i = 0; i < NUMBINS; i++) {
		if (hist[comp][i] > histmax)
			histmax = hist[comp][i];
		}
	    }
	picture_of_int canvas(40,256);
	irange crow(0,1,39);
	irange ccol(0,1,255);
	irange one(0,0,0);
	// Draw axes first
	hists[one+(MAGBLOCK*9+40)][ccol+34] = 255;
	hists.green[one+(MAGBLOCK*9+40+60)][ccol+34] = 255;
	hists.red[one+(MAGBLOCK*9+40+120)][ccol+34] = 255;
	hists.blue[one+(MAGBLOCK*9+40+180)][ccol+34] = 255;
	canvas.draw_graph(hist[0],NUMBINS,255,0,histmax);
	hists[crow+(MAGBLOCK*9)][ccol+34] = canvas;
	canvas = 0;
	canvas.draw_graph(hist[1],NUMBINS,255,0,histmax);
	hists.green[crow+(MAGBLOCK*9+60)][ccol+34] = canvas;
	canvas = 0;
	canvas.draw_graph(hist[2],NUMBINS,255,0,histmax);
	hists.red[crow+(MAGBLOCK*9+120)][ccol+34] = canvas;
	canvas = 0;
	canvas.draw_graph(hist[3],NUMBINS,255,0,histmax);
	hists.blue[crow+(MAGBLOCK*9+180)][ccol+34] = canvas;
	char buf[80];
	sprintf(buf,"%4d",inmin);
	picture_of_int lowval(buf,255,0,1);
	irange minrow(0,1,lowval.nrows()-1);
	irange mincol(0,1,lowval.ncols()-1);
	hists[minrow+(MAGBLOCK*9)+30][mincol+4] = lowval;
	sprintf(buf,"%4d",inmax);
	picture_of_int highval(buf,255,0,1);
	irange maxrow(0,1,highval.nrows()-1);
	irange maxcol(0,1,highval.ncols()-1);
	hists[maxrow+(MAGBLOCK*9)+30][maxcol+290] = highval;
	}

// Put the pel values on the coloured blocks in the magnifying glass
void labelblocks(colour_picture& mag, int ismono, int centx, int centy) {
	char buf[256];
	int i, j;
	centx -= 3;
	centy -= 3;
	// First black out the legend area
	irange legbig(0,1,(MAGBLOCK*8)-1);
	irange legsmall(0,1,MAGBLOCK-1);
	mag[legbig][legsmall] = 0;
	mag[legsmall][legbig] = 0;
	for (i = 0; i < 8; i++) 
	for (j = 0; j < 8; j++) {
		if (!i && !j)
			continue;
		int y = i*MAGBLOCK + MAGBLOCK/2;
		int x = j*MAGBLOCK + MAGBLOCK/2;
		if (!i)
			sprintf(buf,"%d",centx++);
		else if (!j)
			sprintf(buf,"%d",centy++);
		else if (ismono)
			sprintf(buf,"%d",mag.green[y][x]);
		else
			sprintf(buf,"%d\n%d\n%d",mag.red[y][x],
						mag.green[y][x],mag.blue[y][x]);
		picture_of_int label(buf,300,-1,1);
		x -= label.ncols()/2;
		y -= label.nrows()/2;
		irange rr(0,1,label.nrows()-1);
		irange cr(0,1,label.ncols()-1);
		mag.green[rr+y][cr+x] &= label;
		mag.red[rr+y][cr+x] &= label;
		mag.blue[rr+y][cr+x] &= label;
		}
	}
	
// Work out how magnified or diminished the picture on display is
int getzoomlevel(const colour_picture *inpic, const colour_picture *disp) {
	int zoomlevel = 16;
	int scale;
	int nc = inpic->ncols();
	int dispnc = disp->ncols();
	if (nc >= dispnc) {
		scale = nc/dispnc;
		zoomlevel = 17;
		while (scale) {
			zoomlevel--;
			scale >>= 1;
			}
		}
	else {
		scale = dispnc/nc;
		zoomlevel = 15;
		while (scale) {
			zoomlevel++;
			scale >>= 1;
			}
		}
	return zoomlevel;
	}

// mapxy maps from zoomed pointer vals to original (x,y), returning 1 if the
// cursor has moved
int mapxy(colour_picture *inpic, colour_picture *disp, int&x, int&y) {
	int zoomlevel = getzoomlevel(inpic,disp);
	int tempx = disp->pointx();
	int tempy = disp->pointy();
	tempy -= 16;
	if (zoomlevel > 16) {
		zoomlevel -= 16;
		tempx >>= zoomlevel;
		tempy >>= zoomlevel;
		}
	else if (zoomlevel < 16) {
		zoomlevel = 16 - zoomlevel;
		tempx <<= zoomlevel;
		tempy <<= zoomlevel;
		tempx += (1<<(zoomlevel-1));
		tempy += (1<<(zoomlevel-1));
		}
	if ((x == tempx)&&(y == tempy))
		return 0;
	x = tempx;
	y = tempy;
	return 1;
	}

// Draw the control 'menu' onto the top of the displayed picture
void controls(colour_picture *disp, const int channels, const int insp) {
	picture_of_int legend("|+|-|o|i|",255,0,1);
	picture_of_int   legi("       i ",127,0,1);
	if (!insp)
		legend -= legi;
	picture_of_int extleg("r|g|b|",128,0,1);
	picture_of_int extdiv(" | | |",127,0,1);
	extleg += extdiv;
	picture_of_int   extr("r     ",127,0,1);
	picture_of_int   extg("  g   ",127,0,1);
	picture_of_int   extb("    b ",127,0,1);
	irange rleg(2,3,2+legend.nrows()-1);
	irange cleg(0,1,legend.ncols()-1);
	irange cext(legend.ncols(),legend.ncols()+1,
					legend.ncols()+extleg.ncols()-1);
	irange rdisp(0,1,15);
	irange cdisp(0,1,disp->ncols()-1);
	(*disp)[rdisp][cdisp].red = 0;
	(*disp)[rdisp][cdisp].green = 0;
	(*disp)[rdisp][cdisp].blue = 0;
	(*disp)[rleg][cleg].red = legend;
	(*disp)[rleg][cleg].green = legend;
	(*disp)[rleg][cleg].blue = legend;
	if ((disp->ismonochrome())||(disp->ncols() < cext.iend()))
		return;
	(*disp)[rleg][cext].red = extleg;
	(*disp)[rleg][cext].green = extleg;
	(*disp)[rleg][cext].blue = extleg;
	if (channels & 4)
		(*disp)[rleg][cext].red += extr;
	if (channels & 2)
		(*disp)[rleg][cext].green += extg;
	if (channels & 1)
		(*disp)[rleg][cext].blue += extb;
	}
	

// Change the scale of the displayed picture (by factors of 2 in rows&cols
void zoom(colour_picture *inpic, colour_picture **disp, const char *name, 
			int zoomlevel, const int channels, const int insp) {
	bool reshow = 0;
	int nr = inpic->nrows();
	int nc = inpic->ncols();
	int dispnr = nr;
	int dispnc = nc;
	int spacing, lastr, lastc;
	if (zoomlevel == -1) {	// Just a reshow
		reshow = 1;
		zoomlevel = getzoomlevel(inpic,*disp);
		}
	else {
		// Get rid of previously displayed version
		if (*disp) {
			(*disp)->unshow();
			delete *disp;
			}
		}
	if (zoomlevel < 16) {
		dispnr >>= (16-zoomlevel);
		dispnc >>= (16-zoomlevel);
		if (dispnr < 1) dispnr = 1;
		if (dispnc < 1) dispnc = 1;
		spacing = 1<<(16-zoomlevel);
		lastr = (dispnr - 1)*spacing;
		lastc = (dispnc - 1)*spacing;
		irange rr(0,spacing,lastr);
		irange cr(0,spacing,lastc);
		irange outrr(16,17,dispnr+15);
		irange outcr(0,1,dispnc-1);
		if (!reshow)
			*disp = new colour_picture(dispnr+16,dispnc,32,inpic->ismonochrome());
		(**disp)[outrr][outcr] = (*inpic)[rr][cr];
		}
	else if (zoomlevel > 16) {
		dispnr <<= (zoomlevel-16);
		dispnc <<= (zoomlevel-16);
		spacing = 1<<(zoomlevel-16);
		irange block(0,1,spacing-1);
		if (!reshow)
			*disp = new colour_picture(dispnr+16,dispnc,32,inpic->ismonochrome());
		for (int i = 0, ismall = 0; i < dispnr; i += spacing, ismall++)
		for (int j = 0, jsmall = 0; j < dispnc; j += spacing, jsmall++) {
			if (inpic->ismonochrome())
				(*disp)->mono[block+(i+16)][block+j] = inpic->mono[ismall][jsmall];
			else {
				(*disp)->red[block+(i+16)][block+j] = inpic->red[ismall][jsmall];
				(*disp)->green[block+(i+16)][block+j] = inpic->green[ismall][jsmall];
				(*disp)->blue[block+(i+16)][block+j] = inpic->blue[ismall][jsmall];
				}
			}
		}
	else if (!reshow) {
		irange outrr(16,17,dispnr+15);
		irange outcr(0,1,dispnc-1);
		*disp = new colour_picture(dispnr+16,dispnc,32,inpic->ismonochrome());
		(**disp)[outrr][outcr] = *inpic;
		}
	else {
		irange outrr(16,17,dispnr+15);
		irange outcr(0,1,dispnc-1);
		(**disp)[outrr][outcr] = *inpic;
		}
	if (!(*disp)->ismonochrome() && (channels != 7)) {
		if (!(channels & 4))
			(*disp)->red = 0;
		if (!(channels & 2))
			(*disp)->green = 0;
		if (!(channels & 1))
			(*disp)->blue = 0;
		}
	controls(*disp,channels,insp);
	if (!reshow)
		(*disp)->show(name);
	else
		(*disp)->reshow();
	}
		
int main(int argc, char *argv[]) {
	int fromcam = 0;
	if (argc < 2) {
		fromcam = 1;
		argc++;
		cerr << "Usual usage: di picfile1 picfile2 ... Because called without arguments, using camera\n";
		}
	char header[256];
	colour_picture **pp = new colour_picture *[argc];
	colour_picture **ppshow = new colour_picture *[argc];
	int *channels = new int[argc];
	int *inspector = new int[argc];
	for (int picnum = 0; picnum < argc; picnum++) {
		ppshow[picnum] = 0;
		channels[picnum] = 7;
		inspector[picnum] = 0;
		}
	colour_picture *backup = 0;
	int *shown = new int[argc];
	int numshown = argc-1;
	if (fromcam) {
		pp[1] = new colour_picture(480,640);
		zoom(pp[1],ppshow+1,"Camera",16,channels[1],inspector[1]);
		shown[1] = 1;
		pp[1]->attach_camera();
		pp[1]->request_frame();
		}
	else for (int i = 1; i < argc; i++) {
		pp[i] = new colour_picture(argv[i]);
		sprintf(header,"%s %dx%d",argv[i],pp[i]->ncols(), 
							pp[i]->nrows());
		if (pp[i]->ismonochrome())
			strcat(header," monochrome");
		else
			strcat(header," colour");
		zoom(pp[i],ppshow+i,header,16,channels[i],inspector[i]);
		shown[i] = 1;
		}
	int currpic = 0;
	int rcurrpic = 0;
	int inspectpic = 0;
	colour_picture *pic = 0;
	colour_picture *inpic = 0;
	int leftbutton = 0;
	int rightbutton = 0;
	int x, y;
	while (numshown) {
	    if (inspectpic) {	// Currently inspecting a picture
		if (mapxy(inpic,pic,x,y)) {
			if ((x < 0)||(x >= inpic->ncols()-1)||(y < 0)||(y >= inpic->nrows()-1))
				sprintf(header,"%s Pointer outside",argv[inspectpic]);
			else if (inpic->ismonochrome()) {
				sprintf(header,"%s (%d,%d)->%d", argv[inspectpic],x,y,inpic->mono[y][x]);
				int inr, inc, outr, outc, rc, cc;
				for (inr=y-3,rc=MAGBLOCK;inr<=y+3;inr++,rc+=MAGBLOCK) {
				    for (outr = rc; outr < rc+MAGBLOCK; outr++) {
					for (inc=x-3,cc=MAGBLOCK;inc<=x+3;inc++,cc+=MAGBLOCK){
					    for (outc=cc; outc<cc+MAGBLOCK; outc++) {
						backup->green[outr][outc] = inpic->mono[inr][inc];
						backup->red[outr][outc] = inpic->mono[inr][inc];
						backup->blue[outr][outc] = inpic->mono[inr][inc];
						}
					    }
					}
				    }
				}
			else {
				sprintf(header,"%s (%d,%d)->(%d,%d,%d)", argv[inspectpic],x,y,inpic->red[y][x],inpic->green[y][x],inpic->blue[y][x]);
				int inr, inc, outr, outc, rc, cc;
				for (inr=y-3,rc=MAGBLOCK;inr<=y+3;inr++,rc+=MAGBLOCK) {
				    for (outr = rc; outr < rc+MAGBLOCK; outr++) {
					for (inc=x-3,cc=MAGBLOCK;inc<=x+3;inc++,cc+=MAGBLOCK){
					    for (outc=cc; outc<cc+MAGBLOCK; outc++) {
						backup->red[outr][outc] = inpic->red[inr][inc];
						backup->green[outr][outc] = inpic->green[inr][inc];
						backup->blue[outr][outc] = inpic->blue[inr][inc];
						}
					    }
					}
				    }
				}
			labelblocks(*backup,inpic->ismonochrome(),x,y);
			backup->reshow();
			pic->rename(header);
			}
		}
	    // Then scan for button down
	    for (int i = 1; i < argc; i++) {
		if (!shown[i])
			continue;
		if (ppshow[i]->pointleftbutton()) {
			currpic = i;
			leftbutton = 1;
			break;
			}
		else if (leftbutton && (i == currpic)) {
			leftbutton = 0;
			if (ppshow[i]->pointy() < 16) {
				int zl = getzoomlevel(pp[i],ppshow[i]);
				if (ppshow[i]->pointx() < 18)
					zl++;
				else if (ppshow[i]->pointx() < 32)
					zl--;
				else if (ppshow[i]->pointx() < 46) {
					if (zl == 16)
						zl = -1;
					else
						zl = 16;
					}
				else if (ppshow[i]->pointx() < 60) {
					zl = -1;
					if (inspector[i]) {
						backup->unshow();
						delete backup;
						backup = 0;
						inspectpic = 0;
						inspector[i] = 0;
						}
					else {
						for (int w = 1; w < argc; w++){
							if (inspector[w]) {
								inspector[w] = 0;
								zoom(pp[w],ppshow+w,argv[w],zl,channels[w],inspector[w]);
								}
							}
						inspector[i] = 1;
						sprintf(header,"Pel inspector for %s",argv[i]);
						if (!backup) {
							backup = new colour_picture(8*MAGBLOCK+300,8*MAGBLOCK,0);
							makehists(*(pp[i]),*backup);
							backup->show(header);
							}
						else {
							makehists(*(pp[i]),*backup);
							backup->rename(header);
							}
						inspectpic = i;
						pic = ppshow[i];
						inpic = pp[i];
						x = ppshow[i]->pointx();
						y = ppshow[i]->pointy();
						}
					}
				else if (ppshow[i]->pointx() < 74) {
					channels[i] ^= 4; zl = -1; }
				else if (ppshow[i]->pointx() < 88) {
					channels[i] ^= 2; zl = -1; }
				else if (ppshow[i]->pointx() < 100) {
					channels[i] ^= 1; zl = -1; }
				else
					zl = -1;
				zoom(pp[i],ppshow+i,argv[i],zl,channels[i],
						inspector[i]);
				}
			}
		if (ppshow[i]->pointrightbutton()) {
			rcurrpic = i;
			rightbutton = 1;
			break;
			}
		else if (rightbutton&&(i == rcurrpic)) {
			rightbutton = 0;
			if (!fromcam && ppshow[rcurrpic]->key()) {	// If a key is held at the same time as button released
				// Close all
				for (int i = 1; i < argc; i++) {
					if (shown[i])
						ppshow[i]->unshow();
					delete pp[i];
					delete ppshow[i];
				}
				if (inspectpic) {
					backup->unshow();
					delete backup;
					}
				delete [] pp;
				delete [] ppshow;
				return 0;
				}
			ppshow[rcurrpic]->unshow();
			delete ppshow[rcurrpic];
			ppshow[rcurrpic] = 0;
			shown[rcurrpic] = 0;
			numshown--;
			if (rcurrpic == inspectpic) {
				inspector[inspectpic] = 0;
				inspectpic = 0;
				backup->unshow();
				delete backup;
				backup = 0;
				}
			break;
			}
		if (ppshow[i]->closerequested()) {
			ppshow[i]->unshow();
			delete ppshow[i];
			ppshow[i] = 0;
			shown[i] = 0;
			numshown--;
			if (i == inspectpic) {
				inspector[inspectpic] = 0;
				inspectpic = 0;
				backup->unshow();
				delete backup;
				backup = 0;
				}
			break;
			}
		}
	if (inspectpic && backup->closerequested()) {
		inspector[inspectpic] = 0;
		zoom(pp[inspectpic],ppshow+inspectpic,argv[inspectpic],-1,
				channels[inspectpic],inspector[inspectpic]);
		inspectpic = 0;
		backup->unshow();
		delete backup;
		backup = 0;
		}
	if (fromcam) {
		if (pp[1]->frame_delivered()) {
			if (ppshow[1]) {
				zoom(pp[1],ppshow+1,"Camera",-1,channels[1],
						inspector[1]);
				if (inspectpic == 1) {
					makehists(*(pp[1]),*backup);
					x = -1; y = -1;  // Will change insp
							// next time
					}
				pp[1]->request_frame();
				}
			}
		}

	    }
	if (fromcam)
		pp[1]->detach_camera();
	for (int ii = 1; ii < argc; ii++) {
		delete pp[ii];
		if (ppshow[ii])
			delete ppshow[ii];
		}
	delete [] pp;
	delete [] ppshow;
	return 0;
	}
